/* $Id: ByteArrayBuffer.java 544 2008-05-19 20:10:58Z hohwille $
 * Copyright (c) The m-m-m Team, Licensed under the Apache License, Version 2.0
 * http://www.apache.org/licenses/LICENSE-2.0 */
package net.sf.mmm.util.io.impl;

import java.util.NoSuchElementException;

import net.sf.mmm.util.io.api.ByteIterator;

/**
 * This class is similar to {@link java.nio.ByteBuffer} but a lot simpler.
 * However it allows to {@link #setCurrentIndex(int) set the current index} so
 * the internal {@link #getBytes() buffer}-array can be consumed externally and
 * proceeded very fast.<br>
 * <b>ATTENTION:</b><br>
 * This class is NOT intended to be exposed (e.g. by some API). It should only
 * be used internally by some class or component.<br>
 * <b>NOTE:</b><br>
 * This class is NOT public visible, because further releases might break it's
 * compatibility. Feel free to review and give feedback on the mailing list if
 * you want to use it directly.
 * 
 * @see java.nio.ByteBuffer#wrap(byte[], int, int)
 * 
 * @author Joerg Hohwiller (hohwille at users.sourceforge.net)
 */
public class ByteArrayBuffer implements ByteIterator {

  /** @see #getBytes() */
  private final byte[] buffer;

  /** @see #getCurrentIndex() */
  private int currentIndex;

  /** @see #getMaximumIndex() */
  private int maximumIndex;

  /**
   * The constructor.
   * 
   * @param capacity is the <code>length</code> of the internal
   *        {@link #getBytes() buffer}.
   */
  public ByteArrayBuffer(int capacity) {

    this(new byte[capacity], 0, -1);
  }

  /**
   * The constructor.
   * 
   * @param buffer is the internal {@link #getBytes() buffer}.
   */
  public ByteArrayBuffer(byte[] buffer) {

    this(buffer, 0, buffer.length - 1);
  }

  /**
   * The constructor.
   * 
   * @param buffer is the internal {@link #getBytes() buffer}.
   * @param currentIndex is the {@link #getCurrentIndex() current index}.
   * @param maximumIndex is the {@link #getMaximumIndex() maximum index}.
   */
  public ByteArrayBuffer(byte[] buffer, int currentIndex, int maximumIndex) {

    super();
    this.buffer = buffer;
    this.currentIndex = currentIndex;
    this.maximumIndex = maximumIndex;
  }

  /**
   * This method gets the underlying byte-array of this buffer. You are only
   * permitted to read the content from {@link #getCurrentIndex() currentIndex}
   * to {@link #getMaximumIndex() maximumIndex}. Only the creator of this
   * object should modify this array.
   * 
   * @see #getCurrentIndex()
   * @see #getMaximumIndex()
   * 
   * @return the buffer
   */
  public byte[] getBytes() {

    return this.buffer;
  }

  /**
   * This method gets the current index in the {@link #getBytes() buffer}. It
   * will be in the range from <code>0</code> to
   * <code>{@link #getMaximumIndex() maximumIndex} + 1</code>. If this value
   * is less or equal to {@link #getMaximumIndex() maximumIndex}, then
   * <code>{@link #getBytes() buffer}[{@link #getCurrentIndex() currentIndex}]</code>
   * is the next available byte. If it is greater than
   * {@link #getMaximumIndex() maximumIndex}, the {@link #getBytes() buffer} is
   * consumed ({@link #hasNext() has} no next byte).
   * 
   * @return the currentIndex is the current buffer position.
   */
  public int getCurrentIndex() {

    return this.currentIndex;
  }

  /**
   * This method sets the {@link #getCurrentIndex() currentIndex}. This can be
   * done if data from the {@link #getBytes() buffer} has been consumed
   * externally.<br>
   * <b>ATTENTION:</b><br>
   * Be very careful and only use this method if you know what you are doing!
   * 
   * @param currentIndex is the {@link #getCurrentIndex() currentIndex} to set.
   *        It has to be in the range from <code>0</code> to
   *        <code>{@link #getMaximumIndex() maximumIndex} + 1</code>. A value
   *        of <code>{@link #getMaximumIndex() maximumIndex} + 1</code>
   *        indicates that the buffer is consumed.
   */
  public void setCurrentIndex(int currentIndex) {

    assert (currentIndex >= 0);
    assert (currentIndex <= (this.maximumIndex + 1));
    this.currentIndex = currentIndex;
  }

  /**
   * This method gets the maximum index in the {@link #getBytes() buffer}. It
   * will be in the range from <code>-1</code> to
   * <code>{@link #getBytes()}.length - 1</code>.<br>
   * A negative value (<code>-1</code>) indicates that the
   * {@link #getBytes() buffer} does NOT contain data (payload).
   * 
   * @return the maximumIndex
   */
  public int getMaximumIndex() {

    return this.maximumIndex;
  }

  /**
   * @param maximumIndex is the maximumIndex to set
   */
  public void setMaximumIndex(int maximumIndex) {

    this.maximumIndex = maximumIndex;
  }

  /**
   * {@inheritDoc}
   */
  public byte getNext() throws NoSuchElementException {

    if (this.currentIndex > this.maximumIndex) {
      throw new NoSuchElementException();
    }
    return this.buffer[this.currentIndex++];
  }

  /**
   * {@inheritDoc}
   */
  public boolean hasNext() {

    return (this.currentIndex <= this.maximumIndex);
  }

  /**
   * {@inheritDoc}
   */
  public long skip(long byteCount) {

    int bytesLeft = this.maximumIndex - this.currentIndex + 1;
    int skip;
    if (bytesLeft > byteCount) {
      skip = (int) byteCount;
    } else {
      skip = bytesLeft;
    }
    this.currentIndex = this.currentIndex + skip;
    return skip;
  }

  /**
   * This method gets the number of bytes available in this buffer.
   * 
   * @return the bytes left in this buffer.
   */
  public int getBytesAvailable() {

    return this.maximumIndex - this.currentIndex + 1;
  }

}
